package mt.deepnight.slb;

import flash.display.Bitmap;
import flash.display.BitmapData;
import flash.geom.Point;
import flash.geom.Rectangle;
import mt.deepnight.slb.BLib;


class BSprite
extends flash.display.Sprite
{
	public static var UPDATE_LIST : List<BSprite> = new List();

	// Definition
	public var lib(default,null)			: BLib;
	public var group(default,null)			: Null<LibGroup>;
	public var groupName(default,null)		: Null<String>;
	public var frame(default, null)			: Int;

	//exposed for uv stealing
	public var frameData(default,null)		: FrameData;

	// Misc
	var bmp									: Bitmap;
	public var bitmapVisible(default, set)	: Bool;

	var pt0									: flash.geom.Point;
	public var destroyed(default, null )	: Bool;
	public var onFrameChange				: Null<Void->Void>;

	#if debug
	var debug								: flash.display.Sprite;
	#end

	// Anims
	public var a							: AnimManager;

	// Pivots
	// DE : todo optimize this into classes for cpp efficiency
	public var pivotCoord(default,null)		: Null<{x:Float, y:Float}>; // pixel coordinates, excluses pivotCoord
	public var pivotFactor(default,null)	: Null<{xr:Float, yr:Float}>; // homogeneous coordinates : 0-1 (based on actual width/height) excludes pivotFactor

	public function new(?l:BLib, ?g:String, ?frame=0) {
		super();
		destroyed = false;

		bitmapVisible = true;
		this.cacheAsBitmap = false;

		pt0 = new flash.geom.Point();

		a = new AnimManager(this);

		bmp = new Bitmap(flash.display.PixelSnapping.NEVER, false);
		addChild(bmp);

		if( l!=null )
			set(l, g, frame);
	}

	public function clone() {
		var s = new BSprite(lib, groupName, frame);
		s.pivotCoord = pivotCoord;
		s.pivotFactor = pivotFactor;
		s.applyPivot();
		return s;
	}


	public override function toString() {
		return "BSprite_"+groupName+"["+frame+"]";
	}

	public inline function applyFilter(f:flash.filters.BitmapFilter) {
		bmp.bitmapData.applyFilter(bmp.bitmapData, bmp.bitmapData.rect, pt0, f);
	}

	public function set(?l:BLib, ?g:String, ?frame=0, ?stopAllAnims=false) {
		if( l!=null ) {
			lib = l;
			frameData = lib.getFrameData(g, frame);
			if( pivotCoord==null && pivotFactor==null )
				setCenter(lib.defaultCenterX, lib.defaultCenterY);

			if( g==null ) {
				groupName = null;
				group = null;
			}
		}

		if( g!=null )
			groupName = g;

		if( isReady() ) {
			if( stopAllAnims )
				a.stopWithoutStateAnims();
			group = lib.getGroup(groupName);
			initBitmap();
			setFrame(frame);
		}
	}

	public function setRandom(?l:BLib, g:String, rndFunc:Int->Int) {
		set(l, g);
		setFrame(lib.getRandomFrame(groupName, rndFunc));
	}

	function set_bitmapVisible(v) {
		bitmapVisible = v;
		if( bmp!=null )
			bmp.visible = bitmapVisible;
		return v;
	}

	public function getAnimDuration() {
		var a = getAnim();
		return a!=null ? a.length : 0;
	}

	inline function getAnim() {
		return isReady() && group.anim.length>=0 ? group.anim : null;
	}

	inline function hasAnim() {
		return getAnim()!=null;
	}

	public inline function getBitmap() {
		return bmp;
	}

	public inline function getBitmapData() {
		return bmp.bitmapData;
	}

	public inline function isGroup(k) {
		return groupName==k;
	}

	function initBitmap() {
		if( !isReady() )
			return;

		if( bmp.bitmapData==null || group.maxWid>bmp.bitmapData.width || group.maxHei>bmp.bitmapData.height ) {
			var oldW = 0;
			var oldH = 0;
			if( bmp.bitmapData!=null ) {
				//#if debug
				//trace('WARNING($groupName): re-allocation of bitmapdata occured (new size: ${group.maxWid}x${group.maxHei})');
				//#end
				oldW = bmp.bitmapData.width;
				oldH = bmp.bitmapData.height;
				bmp.bitmapData.dispose();
				bmp.bitmapData = null;
			}
			bmp.bitmapData = new BitmapData(MLib.max(oldW, group.maxWid), MLib.max(oldH, group.maxHei), true, 0x0);
		}
		bmp.visible = bitmapVisible;
		applyPivot();
	}

	#if debug
	public function enableDebug() {
		if( debug!=null )
			debug.parent.removeChild(debug);

		debug = new flash.display.Sprite();
		addChild(debug);
	}

	inline function updateDebug() {
		if( debug!=null ) {
			var g = debug.graphics;
			g.clear();

			// Pivot
			g.lineStyle(1,0xFFFF00,0.8, true, flash.display.LineScaleMode.NONE);
			g.moveTo(0,-3);
			g.lineTo(0,3);
			g.moveTo(-3,0);
			g.lineTo(3,0);

			// Real frame bounds
			g.lineStyle(1,0xFF77FF,0.8, true, flash.display.LineScaleMode.NONE);
			g.drawRect(bmp.x+frameData.realFrame.x, bmp.y+frameData.realFrame.y, frameData.realFrame.realWid, frameData.realFrame.realHei);

			// Bitmap bounds
			g.lineStyle(0,0,0);
			g.beginFill(0x00FFFF,0.3);
			g.drawRect(bmp.x, bmp.y, bmp.width, bmp.height);
			g.endFill();

			// Current frame bounds
			g.lineStyle(1,0x00FFFF,0.8, true, flash.display.LineScaleMode.NONE);
			g.drawRect(bmp.x, bmp.y, frameData.wid, frameData.hei);
		}
	}
	#end

	public inline function randomFrame(?rndFunc:Int->Int) {
		if( isReady() ) {
			if( rndFunc==null )
				rndFunc = Std.random;
			setFrame( rndFunc(lib.countFrames(groupName)) );
		}
	}

	public inline function setFrame(f:Int) {
		var old = frame;
		frame = f;
		if( isReady() && (old!=frame || frame==0) ) { // TODO frame 0 pas terrible...
			var prev = frameData;

			frameData = lib.getFrameData(groupName, frame);

			if( frameData.wid<prev.wid || frameData.hei<prev.hei )
				bmp.bitmapData.fillRect( bmp.bitmapData.rect, 0 );

			bmp.bitmapData.copyPixels( lib.source, frameData.rect, pt0,null,null, false );
			//if( frameData.realFrame.x!=prev.realFrame.x || frameData.realFrame.y!=prev.realFrame.y ) {
				//bmp.x += prev.realFrame.x  - frameData.realFrame.x;
				//bmp.y += prev.realFrame.y  - frameData.realFrame.y;
			//}

			applyPivot();
			if( onFrameChange!=null )
				onFrameChange();
		}
	}

	public inline function setPivotCoord(x:Float, y:Float) {
		pivotCoord = {x:x, y:y}
		pivotFactor = null;
		applyPivot();
	}

	public inline function setCenter(xRatio:Float, yRatio:Float) {
		pivotCoord = null;
		pivotFactor = {xr:xRatio, yr:yRatio};
		applyPivot();
	}


	inline function applyPivot() {

		if( pivotCoord!=null ) {
			bmp.x = MLib.round(-pivotCoord.x - frameData.realFrame.x);
			bmp.y = MLib.round(-pivotCoord.y - frameData.realFrame.y);
		}
		else if( pivotFactor!=null ) {
			bmp.x = Std.int(-frameData.realFrame.realWid*pivotFactor.xr - frameData.realFrame.x);
			bmp.y = Std.int(-frameData.realFrame.realHei*pivotFactor.yr - frameData.realFrame.y);
		}

		#if debug
		updateDebug();
		#end
	}


	public function totalFrames() {
		return group.frames.length;
	}

	public function destroy() {
		if( !destroyed ) {
			destroyed = true;

			a.destroy();
			a = null;

			if(parent!=null)
				parent.removeChild(this);

			if( bmp.bitmapData!=null ) {
				bmp.bitmapData.dispose();
				bmp.bitmapData = null;
			}
		}
	}

	public inline function isReady() {
		return !destroyed && groupName!=null;
	}

	public static inline function updateAll() {
		for( a in AnimManager.UPDATE_LIST )
			a.update();
	}


}

